#include "LDSMac.h"

Define_Module(LDSMac);


/*-- FONCTIONS PROPRES A LA SIMULATION --*/
/*Fonction appelée a l'initialisation*/
void LDSMac::startup(){
	trace() << "Initialisation de la couche mac LDSMac";

	//on retrouve le module de gestion de position pour avoir la position du module
	cModule *parent = getParentModule()->getParentModule();//on recupere le module node
	if (parent->findSubmodule("MobilityManager") != -1) {//on cherche le module de gestion de position
		mobilityModule = check_and_cast <VirtualMobilityManager*>(parent->getSubmodule("MobilityManager"));
	} 
	else{
		opp_error("\n[Application]:\n Error in geting a valid reference to MobilityManager for direct method calls.");
	}	
	NodePosition position;//on créé les données
	position.setNode(SELF_MAC_ADDRESS);
	position.setX(mobilityModule->getLocation().x);
	position.setY(mobilityModule->getLocation().y);
	position.setZ(mobilityModule->getLocation().z);
	nodePositionMap[SELF_MAC_ADDRESS] = position;//on stocke notre propre position dans la map
	
	//données provenant du NED a parser
	largeurTubeEnvoi = (float) (int) par("largeurTubeEnvoi");
	useAck = par("useAck");
	sinkNode = par("sinkNode");
	
	//données interne a la couche mac
	currentSource = -1;

	//variable de paquet
	dataPkt = NULL;
	rtsPkt = NULL;
	ctsPkt = NULL;
	ackPkt = NULL;

}

/*Fonction appelée a la fin de la simulation*/
void LDSMac::finishSpecific(){
	trace() << "Fin de la simulation avec la couche mac LDSMac";	
}
/*-******************************************************************************************************************/


/* Fonctions propres a la couche mac utilitaire */
//cette fonction retourne si le noeud courant peut servir de relais au message venant de emeteur vers destination
//bool LDSMac::isInPath(int emeteur,int destination){
//	NodePosition pos_emeteur = nodePositionMap[emeteur];//on recupere la position du noeud emeteur de données
//	NodePosition pos_destination = nodePositionMap[destination];//la position du noeud destination
//	NodePosition pos_noeud_courant = nodePositionMap[SELF_MAC_ADDRESS];//la position du noeud courant	
//	//calcule la hauteur du noeud par rapport au chemin optimal
//	float hauteurNoeud = HauteurTriangle3Node(pos_emeteur,pos_noeud_courant,pos_destination);
//	trace() << "distance a la droite: "<< hauteurNoeud;
//	return (hauteurNoeud < largeurTubeEnvoi);
//}

////DEBUG


bool LDSMac::isInPath(int emeteur,int destination){
	return false;	
}

bool LDSMac::radioIsClear(){
	return (radioModule->isChannelClear() == CLEAR);
}

/*Traite les paquets reçus de la couche de routage*/
void LDSMac::fromNetworkLayer(cPacket *pkt,int dstAddr){ //Paquets sortants
	if(SELF_MAC_ADDRESS != sinkNode){ 
		if(dstAddr == SELF_MAC_ADDRESS){//au cas ou on essaie de s'envoyer le paquet a nous meme 
			toNetworkLayer(pkt);
			return; 
		}
		LDSMacPacket *dataInPkt = new LDSMacPacket("LDSMac DATA Packet",MAC_LAYER_PACKET);//on cree un paquet de donnée que l'on bufferise
		dataInPkt->setType(DATA_PACKET);
		dataInPkt->setEnvoyeur(SELF_MAC_ADDRESS);
		dataInPkt->setEmeteur(SELF_MAC_ADDRESS);
		dataInPkt->setDestinataire(sinkNode);//on parametre le paquet macdata 
		encapsulatePacket(dataInPkt,pkt);//on encapsule les données dans le paquet mac
		bufferPacket(dataInPkt);//on bufferise le paquet
		sendNextPacket();//on essai d'envoyer le paquet
	}
}



/*Traite le buffer afin d'envoyer tous les messages qu'il contient*/
void LDSMac::timerFiredCallback(int signal){
	if(signal == SLEEP_TIMEOUT){
		macState = WAIT_FOR_RTS;
		sendNextPacket();
		return;
	}
	if(macState == OFF) return;
	trace() << "Rappel de fonction";
	switch(signal){
		case SEND_NEXT_PACKET:
			sendNextPacket();
			break;		
		case RTS_TIMEOUT:
			sendRtsPacket();
			break;
		case DATA_TIMEOUT:
			sendDataPacket();
			break;
		case CTS_TIMEOUT:
			sendCtsPacket();
			break;
		case ACK_TIMEOUT:
			sendAckPacket();
			break;
		default:break;
	}
}

void LDSMac::sendNextPacket(){
	if(macState != WAIT_FOR_RTS || TXBuffer.empty()){
		trace() << " - Ne peut retransmettre";
		return;//on sort de la fonction si la couche mac n'est pas en attente ou si il n'y a pas de paquet a envoyer
	}
	if(!radioIsClear()){
		trace() << " - Radio occupee";
		setTimer(SEND_NEXT_PACKET,5.0);//si la couche radio n'est pas libre on relance la fonction dans n pas de temps//5 sec
		return;
	}
	trace() << " - Envoie premier paquet de la file";
	//if(dataPkt != NULL) cancelAndDelete(dataPkt);
	dataPkt = check_and_cast < LDSMacPacket * >(TXBuffer.front());
	if(dataPkt->getDestinataire() == BROADCAST_MAC_ADDRESS){//soit on envoi un paquet broadcast
		trace() << " - broadcast" << BROADCAST_MAC_ADDRESS;
		sendBroadcastPacket();		
	}
	else{//soit un paquet unicast
		trace() << " - unicast";
		sendUnicastPacket();
	}
}

void LDSMac::sendBroadcastPacket(){	
	//TODO: Traiter correctement les paquets broadcast
}
void LDSMac::sendUnicastPacket(){
	//creation du paquet rts
	//if(rtsPkt != NULL) cancelAndDelete(rtsPkt);//si le paquet est deja utilisé on le detruit
	rtsPkt = new LDSMacPacket("LDSMac RTS Packet",MAC_LAYER_PACKET);//on cree un paquet rts que l'on diffuse
	rtsPkt->setType(RTS_PACKET);
	rtsPkt->setEnvoyeur(SELF_MAC_ADDRESS);
	rtsPkt->setEmeteur(SELF_MAC_ADDRESS);
	rtsPkt->setDestinataire(sinkNode);
	
	currentSource = SELF_MAC_ADDRESS;//on met la source a la valeur du noeud lui meme
	sendRtsPacket();//on donne l'ordre d'envoi du rts
}
void LDSMac::sendRtsPacket(){
	if(!radioIsClear()){//si la porteuse n'est pas libre on arme un timer pour reessayer plus tard
		trace() << "Porteuse non disponible";
		setTimer(RTS_TIMEOUT,3.0);
		return;
	}
	trace() << "-> RTS(" << rtsPkt->getEnvoyeur()<<","<< rtsPkt->getEmeteur() <<","<< rtsPkt->getDestinataire() << ")";
	macState = WAIT_FOR_RTS_CTS;//on se met en etat d'attente
	toRadioLayer(rtsPkt);
	toRadioLayer(createRadioCommand(SET_STATE, TX));
	
	//setTimer(RTS_TIMEOUT,3.0);//on arme ce timer afin de ne pas attendre indefiniment*/
}

void LDSMac::sendDataPacket(){	
	if(!radioIsClear()){//si la porteuse n'est pas libre on arme un timer pour reessayer plus tard
			setTimer(DATA_TIMEOUT,2.0);
			return;
	}
	TXBuffer.pop();//on ne supprime le paquet du buffer que au reel envoi de celui ci
	trace() << "-> DATA(" << dataPkt->getEnvoyeur()<<","<< dataPkt->getEmeteur() <<","<< dataPkt->getDestinataire() << ")";
	toRadioLayer(dataPkt);
	toRadioLayer(createRadioCommand(SET_STATE, TX));
	if(useAck){
		macState = WAIT_FOR_ACK;//on attend l'acquitement du paquet
	}
	else{
		macState = WAIT_FOR_RTS;
		currentSource = -1;
	}
}

void LDSMac::sendCtsPacket(){
	if(!radioIsClear()){//si la porteuse n'est pas libre on arme un timer pour reessayer plus tard
		setTimer(CTS_TIMEOUT,2.0);
		return;
	}
	trace() << "-> CTS(" << ctsPkt->getEnvoyeur()<<","<< ctsPkt->getEmeteur() <<","<< ctsPkt->getDestinataire() << ")";
	toRadioLayer(ctsPkt);
	toRadioLayer(createRadioCommand(SET_STATE, TX));
}

void LDSMac::sendAckPacket(){
	if(!radioIsClear()){//si la porteuse n'est pas libre on arme un timer pour reessayer plus tard
		setTimer(ACK_TIMEOUT,1.0);
		return;
	}
	trace() << "-> ACK(" << ackPkt->getEnvoyeur()<<","<< ackPkt->getEmeteur() <<","<< ackPkt->getDestinataire() << ")";
	toRadioLayer(ackPkt);
	toRadioLayer(createRadioCommand(SET_STATE, TX));	
}




/*Traite les paquets reçus de la couche radio*/
void LDSMac::fromRadioLayer(cPacket *pkt,double RSSI,double LQI){ //Paquets rentrants
	LDSMacPacket *macPkt = dynamic_cast <LDSMacPacket*>(pkt);
	if (macPkt == NULL)
		return;
	//on dispatch le paquet en fonction de l'etat de la couche mac
	switch(macState){//selon notre etat on traite le paquet ou non
		case WAIT_FOR_RTS: waitForRts(macPkt); break;
		case WAIT_FOR_RTS_CTS: waitForRtsCts(macPkt); break;
		case WAIT_FOR_CTS: waitForCts(macPkt); break;
		case WAIT_FOR_DATA: waitForData(macPkt); break;
		case WAIT_FOR_ACK: waitForAck(macPkt); break;
		default : break;	
	}	
}

void LDSMac::waitForRts(LDSMacPacket *macPkt){
	trace() << "Dans waitForRts";
	if(macPkt->getType() == RTS_PACKET){
		trace() << "<- RTS(" << macPkt->getEnvoyeur()<<","<< macPkt->getEmeteur() <<","<< macPkt->getDestinataire() << ")";
		if(macPkt->getDestinataire() == SELF_MAC_ADDRESS){//si on est le noeud cible
			//on retourne un cts
			currentSource = macPkt->getEmeteur();
			//if(ctsPkt != NULL) cancelAndDelete(rtsPkt);//si le paquet est deja utilisé on le detruit
			ctsPkt = new LDSMacPacket("LDSMac CTS Packet",MAC_LAYER_PACKET);//on cree un paquet de donnée que l'on bufferise
			ctsPkt->setType(CTS_PACKET);
			ctsPkt->setEnvoyeur(SELF_MAC_ADDRESS);
			ctsPkt->setEmeteur(macPkt->getEmeteur());
			ctsPkt->setDestinataire(macPkt->getDestinataire());
			macState = WAIT_FOR_DATA;
			sendCtsPacket();
		}	
		else{
			if(isInPath(macPkt->getEmeteur(),macPkt->getDestinataire())){//soit on est un noeud du chemin
				//on reemet le rts avec les nouvelles données 
				rtsPkt = new LDSMacPacket("LDSMac RTS Packet",MAC_LAYER_PACKET);//on cree un paquet rts que l'on diffuse
				rtsPkt->setType(RTS_PACKET);
				rtsPkt->setEnvoyeur(SELF_MAC_ADDRESS);
				rtsPkt->setEmeteur(macPkt->getEmeteur());
				rtsPkt->setDestinataire(macPkt->getDestinataire());
				macState = WAIT_FOR_RTS_CTS;
				sendRtsPacket();
			}
			else{//soit on ne l'est pas
				trace() << "Extinction du noeud";
				macState = OFF;
				setTimer(SLEEP_TIMEOUT,3.0);
			
			}
		}	
	}	
	else{
		//le paquet n'est pas du type que l'on attend, donc une communication est en cours. On eteind le noeud
		trace() << "Extinction du noeud";
		macState = OFF;
		setTimer(SLEEP_TIMEOUT,3.0);
	}
}
void LDSMac::waitForRtsCts(LDSMacPacket *macPkt){
	trace() << "Dans waitForRtsCts";
	if(macPkt->getType() == RTS_PACKET){
		trace() << "<- RTS(" << macPkt->getEnvoyeur()<<","<< macPkt->getEmeteur() <<","<< macPkt->getDestinataire() << ")";		
	}
	else if(macPkt->getType() == CTS_PACKET){
		trace() << "<- CTS(" << macPkt->getEnvoyeur()<<","<< macPkt->getEmeteur() <<","<< macPkt->getDestinataire() << ")";
		if(macPkt->getEmeteur() == currentSource){//cela signifie que le CTS nous est bien destiné
			cancelTimer(RTS_TIMEOUT);//On annule le timer d'attente de CTS
			macState = SEND_DATA;
			sendDataPacket();
		}
		else{
			cancelTimer(RTS_TIMEOUT);//On annule le timer d'attente de CTS
			trace() << "Extinction du noeud";
			macState = OFF;
			setTimer(SLEEP_TIMEOUT,3.0);
		}
	}
}
void LDSMac::waitForCts(LDSMacPacket *macPkt){
	trace() << "Dans waitForCts";
	if(macPkt->getType() == CTS_PACKET){
		trace() << "<- CTS(" << macPkt->getEnvoyeur()<<","<< macPkt->getEmeteur() <<","<< macPkt->getDestinataire() << ")";
		if(macPkt->getEmeteur() == currentSource){//cela signifie que le CTS nous est bien destiné
			cancelTimer(RTS_TIMEOUT);//On annule le timer d'attente de CTS
			macState = SEND_DATA;
			sendDataPacket();
		}
		else{//si le cts ne nous est pas destiné on eteind le noeud pour le moment
			trace() << "Extinction du noeud";
			macState = OFF;
			setTimer(SLEEP_TIMEOUT,3.0);
		}
	}
}

void LDSMac::waitForData(LDSMacPacket *macPkt){
	trace() << "Dans waitForData";
	if(macPkt->getType() == DATA_PACKET){
		trace() << "<- DATA(" << macPkt->getEnvoyeur()<<","<< macPkt->getEmeteur() <<","<< macPkt->getDestinataire() << ")";
		if(macPkt->getEnvoyeur() == currentSource){//si on attend bien un paquet de cette source
			toNetworkLayer(macPkt->decapsulate());//on retourne le paquet a la couche routage
			macState = WAIT_FOR_RTS;
			if(useAck){//si on utilise le ack alors on l'envoi
				if(ackPkt != NULL) cancelAndDelete(rtsPkt);//si le paquet est deja utilisé on le detruit
				ackPkt = new LDSMacPacket("LDSMac ACK Packet",MAC_LAYER_PACKET);//on cree un paquet de donnée que l'on bufferise
				ackPkt->setType(ACK_PACKET);
				ackPkt->setEnvoyeur(SELF_MAC_ADDRESS);
				ackPkt->setEmeteur(macPkt->getEmeteur());
				ackPkt->setDestinataire(macPkt->getDestinataire());
				sendAckPacket();
			}
		}
	}
}
void LDSMac::waitForAck(LDSMacPacket *macPkt){
	trace() << "Dans waitForAck";
	if(macPkt->getType() == ACK_PACKET){
		trace() << "<- ACK(" << macPkt->getEnvoyeur()<<","<< macPkt->getEmeteur() <<","<< macPkt->getDestinataire() << ")";
		if(macPkt->getEmeteur() == currentSource){//si on attend bien un paquet de cette source
			macState = WAIT_FOR_RTS;
			currentSource = -1;	
			sendNextPacket();
		}
	}
		
}

int LDSMac::handleRadioControlMessage(cMessage * msg)
{
	RadioControlMessage *radioMsg = check_and_cast <RadioControlMessage*>(msg);
	toNetworkLayer(msg);
	return 1;
}
