#ifndef _LDSMAC_H_
#define _LDSMAC_H_

#include "VirtualMac.h"
#include "VirtualMobilityManager.h"
#include "LDSMacPacket_m.h"
#include "LDSUtils.h"
#include <map>


enum LDSMacEtat {
	WAIT_FOR_RTS,
	WAIT_FOR_RTS_CTS,
	WAIT_FOR_CTS,
	WAIT_FOR_DATA,
	WAIT_FOR_ACK,
	OFF,
	SEND_DATA
};

enum LDSTimerSignal {
	SEND_NEXT_PACKET,
	DATA_TIMEOUT,
	RTS_TIMEOUT,
	CTS_TIMEOUT,
	ACK_TIMEOUT,
	SLEEP_TIMEOUT
};


class LDSMac : public VirtualMac{
	
	protected:
		
		static map<int,NodePosition> nodePositionMap;
		//cPacket* ackWaitingPacket;
		LDSMacEtat					 macState;
		
				
		VirtualMobilityManager *mobilityModule;	//a pointer to the mobilityModule object
		
		//Variables NED
		float 	largeurTubeEnvoi;
		bool	useAck;
		int 	sinkNode;
		
		//Variables internes
		int 	currentSource;
		
		//paquet créés car il peuvent ne pas etre envoyer tout de suite
		LDSMacPacket	*dataPkt;
		LDSMacPacket	*rtsPkt;
		LDSMacPacket	*ctsPkt;
		LDSMacPacket	*ackPkt;
		
		
	/*-- FONCTIONS PROPRES A LA SIMULATION --*/
		/*Fonction appelée a l'initialisation*/
		void startup();
		/*Fonction appelée a la fin de la simulation*/
		void finishSpecific();
		
		/*Fonction qui réagit aux commandes de controle de la couche radio*/
		int handleRadioControlMessage(cMessage *msg);
		
	/*-- FONCTIONS PROPRES AU MODULE MAC --*/
		/*callback du timer*/
		void timerFiredCallback(int signal);
		/*retourne vrai si le noeud sert de relai sur le chemin entre l'emeteur et la destination*/
		bool isInPath(int emeteur,int destination);
		/*retourne true si la porteuse est libre et false sinon*/
		bool radioIsClear();

		/*Traite les paquets reçus de la couche de routage*/
		void fromNetworkLayer(cPacket *pkt,int dstAddr); //Paquets sortants
		void sendNextPacket();//essai d'envoyer le prochain paquet du buffer avec toute la procedure
		void sendBroadcastPacket();
		void sendUnicastPacket();
		void sendDataPacket();//envoie le premier paquet du buffer sans le supprimer des que la porteuse est libre
		void sendRtsPacket();//envoie le paquet rts courant des que la porteuse est libre
		void sendCtsPacket();//envoie le paquet cts courant des que la porteuse est libre
		void sendAckPacket();//envoie le paquet ack courant des que la porteuse est libre
		
		
		/*Traite les paquets reçus de la couche radio*/
		void fromRadioLayer(cPacket *pkt,double RSSI,double LQI); //Paquets rentrants
		//Fonctions de dispatch chacune d'elle est appelé en fonction de l'etat de la couche mac
		void waitForRts(LDSMacPacket *macPkt);
		void waitForRtsCts(LDSMacPacket *macPkt);
		void waitForCts(LDSMacPacket *macPkt);
		void waitForData(LDSMacPacket *macPkt);
		void waitForAck(LDSMacPacket *macPkt);	
};

map<int,NodePosition> LDSMac::nodePositionMap;


#endif
